# +
#  NAME:
# 
#   QUICKCV
# 
#  PURPOSE:
# 
#    calculate cosmic variance for some geometry
# 
#  CATEGORY:
# 
#    cosmic variance
# 
# 
#  CALLING SEQUENCE:
# 
#    fracerror=quickcv(side1,side2 [zarr,deltaz,sig8=sig8,npoints=npoints]) 
# 
#  INPUTS:
# 
#    side1: length of 1 side of field on sky, in degrees (rect. geom)
#    side2: length of other side of field on sky, in degrees
# 
#  OPTIONAL INPUTS:
#    zarr: array of z's for which cosmic variance will be calculated default:1
#    deltaz: total length of volume in z direction default: 0.1
# 
#  KEYWORD PARAMETERS:
#    sig8: desired value of sigma_8 default: 0.77
#    npoints: desired number of integration points for first part of 
#      integration default: 200
# 
# 
#  OUTPUTS:
# 
#   fracerror: array containing the fractional error in a count,
#   sigma/N, due to cosmic variance for objects with bias b=1 
#   in a volume side1 x side2 degrees x delta_z centered at redshifts
#   zarr.  NOTE THIS IS SIGMA, NOT VARIANCE!!!!
# 
#  RESTRICTIONS:
# 
#   uses power spectrum in pofk.pro (gamma=0.1872, omega0=0.26), distance
#   relation in rz.pro 
# 
# 
#  MODIFICATION HISTORY:
#  modified by rss july 2006 to include Dlin scaling with z
#  modified by bpm sep  2009 to wmap3 cosmology
# -

import numpy as np
from .utils.rz import rz
from .utils.intpk4 import intpk4
from .utils.dlin import Dlin

def quickcv(side1, side2, za, deltaz, OmegaM = 0.31, OmegaL = 0.69, OmegaBaryon = 0.022/(0.678)**2,sigma8 = 0.82, ns = 0.96, h = 0.678, verbose = False, acc='low'):
  """za and deltaz have to be np.arrays"""
  #sides in degrees, z is redshift bin middle, deltaz is redshift bin width, everything else is cosmology and integration parameters
  #see intpk4.py for acc options

  # assume LCDM

  rza=rz(za, OmegaM=OmegaM, OmegaL=OmegaL)
  rzmin=rz(za-deltaz/2, OmegaM = OmegaM, OmegaL = OmegaL)
  rzmax=rz(za+deltaz/2, OmegaM = OmegaM, OmegaL = OmegaL)

  radeg = 57.2958 # degrees per radians

  x1a=3000./2.*(rza * side1)/radeg
  x2a=3000./2.*(rza * side2)/radeg
  x3a=3000.*(rzmax-rzmin)/2.
  xs = np.array([x1a, x2a, x3a])
  cvi = intpk4(xs, OmegaM = OmegaM, OmegaL = OmegaL, OmegaBaryon = OmegaBaryon, sigma8 = sigma8, ns = ns, h = h, acc =  acc, verbose =  verbose)

  #cvi is the fractional variance in a count in a rectangle
  #cvar is the fractional error (sigma), NOT VARIANCE

  cvi = cvi*sigma8**2 #normalize by sig8^2
  d = Dlin(OmegaM, za)
  cvar = np.sqrt(cvi)*d

  return cvar[0]

# print(quickcv(1,2, np.array([3]), np.array([1])))


from .pofk import pofkint
import numpy as np
from scipy.integrate import nquad
import time

def intpk4(xs,  OmegaM = 0.308, OmegaBaryon = 0.022/(0.678)**2, sigma8 = 0.82, ns = 0.96, h = 0.678, acc = 'high', verbose = False):

	"""" Solver for integral in equation 2 of https://arxiv.org/pdf/astro-ph/0109130.pdf
	xs are the side lengths of the box in comoving Mpc
	acc is the accuracy of the integral, 'high' or 'low', low is about 40 times faster and has about a 2% error
	verbose is whether to print out the time taken for the integral

	This version is written by Christian Kragh Jespersen, January 2023, rewritten from the IDL code by Jonathan Baker
	"""
	if acc == 'high':
		acc = 1
	elif acc == 'low':
		acc = 5

	# kmax = 4*np.pi/xs ## this is the maximum k value for the window function that contributes significantly to the integral

	kmax = (0.4+16/xs)*1.25 #this is to match from the idl code
	kmax = kmax/acc

	def window_idl(k0, k1, k2):
		return np.sin(k0*xs[0])/(k0*xs[0])*np.sin(k1*xs[1])/(k1*xs[1])*np.sin(k2*xs[2])/(k2*xs[2]) 

	def integrand_idl(k0,k1,k2):
		return window_idl(k0, k1, k2)**2*pofkint(k0,k1,k2,  OmegaM=OmegaM, OmegaBaryon=OmegaBaryon, sigma8=sigma8, ns = ns, h = h)

	start = time.time()

	integral = nquad(integrand_idl, [[0, kmax[0]], [0, kmax[1]], [0, kmax[2]]], opts = {"limlst": 10, "maxp1": 10})
	end = time.time()

	if verbose:
		print(f'Integration time {end - start:.2f} seconds')

	## factor of pi**3 is to normalize volume in k space, there are offsetting factors of 8 not included here
	return integral[0]/np.pi**3





# Routines for computing the Alcock-Paczynski distortion

# January 2023, Christian Kragh Jespersen, rewritten from IDL


# H(z) in units where H(0) = 1

import numpy as np 
from scipy.integrate import quad

#get hubble parameter at redshift z relative to h at z=0
def hubble(z, OmegaM = 0.308, OmegaL = 0.692):
    x = 1. + z
    OmegaK = 1. - OmegaM - OmegaL #curvature.
    h2 = OmegaM * x**3 + OmegaK * x**2 + OmegaL
    return np.sqrt(h2)

# Compute d(chi)/dz where d(chi)^2 = dr^2 / (1-kr^2)
def dchi_dz(z):
    return 1./hubble(z)

# Compute coordinate distance r(z) for FRW metric
def rz(z, OmegaM = 0.308, OmegaL = 0.692, eps = 1e-10):
    '''Compute coordinate distance r(z) for FRW metric, given cosmological parameters OmegaM and OmegaL and redshift z
    eps parameter is the precision for scipy.integrate.quad'''
    kurv = OmegaM + OmegaL - 1.
    nz = len(z)
    dchi = np.zeros(nz)
    chi = np.zeros(nz)
    r = np.zeros(nz)
    z2 = z[0]
    if z2 == 0.:
        dchi[0] = 0.
    else:
        dchi[0] = quad(dchi_dz, 0., z2, epsrel=eps)[0]
    for i in range(1, nz):
        z1 = z[i-1]
        z2 = z[i]
        dchi[i] = quad(dchi_dz, z1, z2, epsrel=eps)[0]
    chi[0] = dchi[0]
    for i in range(1, nz):
        chi[i] = chi[i-1] + dchi[i]
    if abs(kurv) < 1.e-4: #flat
        r = chi
    elif kurv > 0.: #closed
        r = np.sin(chi*np.sqrt(kurv))/np.sqrt(kurv)
    else: #open
        r = np.sinh(chi*np.sqrt(-kurv))/np.sqrt(-kurv)
    return r


# Rewritten by Christian Kragh Jespersen, @astrockragh, to work in Python 3.8 (Jan-2023)
# Dependencies are numpy, scipy, and pandas
# If anyone (including future me) wants to optimize for speed, the bottleneck is the intpk4 integral


import numpy as np
import pandas as pd
from .quickcv import quickcv
import os

#### Sample parameter ####

# ########### Input parameters ###########
# s1 = 3.4/60 #side 1, in degrees 
# s2 = 10.2/60 #side 2, in degrees
# zarr = np.array([3,4,5,6.5]) #redshift bin edges
# ############################################

# ########### Cosmology ###########
# OmegaM = 0.308
# OmegaL = 0.692
# sigma8 = 0.82
# acc = 'low' # low takes about 20 seconds per calculation, high takes about 8 minutes per calculation
############################################

# The bias values for 7.0, 7.5, and 8.0 were incorrectly extrapolated, here they are set to be the same as for 8.5 in this version by astrockragh 19/11-2022

b070=0.062
b170=2.59
b270=1.025

b075=0.062
b175=2.59
b275=1.025

b080=0.062
b180=2.59
b280=1.025

b085=0.062
b185=2.59
b285=1.025

b090=0.074
b190=2.58
b290=1.039

b095=0.042
b195=3.17
b295=1.147

b0100=0.053
b1100=3.07
b2100=1.225

b0105=0.069
b1105=3.19
b2105=1.269

b0110=0.173
b1110=2.89
b2110=1.438

def get_cv(side1, side2, zarr, name = None, dz = None, acc = 'low', OmegaM = 0.308, OmegaL = 0.692, OmegaBaryon = 0.022/(0.678)**2, sigma8 = 0.82, ns = 0.96, h = 0.678, verbose = False):
   ''' Method to calculate the cosmic variance for a given survey area and redshift bins.
      Give side lengths in degrees
      side1: side length 1 [degrees]
      side2: side length 2 [degrees]
      zarr: array of redshift bin edges
      name: name of the output file. File will be saved at dfs/name.csv if name is not None
      dz: redshift bin size. If None, the redshift bin size will be the difference between the redshift bin edges. If dz is given, zarr is the redshift bin centers.
      acc: accuracy of the calculation. 'low' or 'high'. 'low' is faster (20 sec per calculation), 'high' is more accurate (8 mins per calculation)
      OmegaM, OmegaL, OmegaBaryon, sigma8, ns, h: cosmological parameters
      verbose: whether to print out the time taken for the integral
      
      Output is dataframe with columns:
      zmid: redshift bin center
      dz: redshift bin size
      cv_dm: cosmic variance for dark matter
      cv_xx: cosmic variance for x.x - 0.25 < log(M_*/M_{sun}) < x.x + 0.25

      if given name, the dataframe will be saved to dfs/name.csv, along with a df with metadata, dfs/name_meta.csv so that the runs don't have to be repeated
      '''

   metadict = {'side1': side1,
            'side2': side2,
            'zarr': np.array(zarr),
            'dz': dz,
            'acc': acc,
            'OmegaM': OmegaM,
            'OmegaL': OmegaL,
            'OmegaBaryon': OmegaBaryon, 
            'sigma8': sigma8,
            'ns': ns,
            'h': h,
            'verbose': verbose}

   df_meta = pd.DataFrame(metadict)
   if dz is None:
      nz = len(zarr)-1
   else:
      nz = len(zarr)

   columns = ['zmid', 'dz', 'cv_dm', 'cv_70', "cv_75", "cv_80", "cv_85", "cv_90", "cv_95", "cv_100", "cv_105", "cv_110"]
   df_values = pd.DataFrame(index = np.arange(nz), columns = columns)

   for i in range(nz):
      if dz is None:
         dz      = zarr[i+1]-zarr[i]
         zmid    = zarr[i+1]+zarr[i]
         zmid    = zmid/2
      else:
         zmid    = zarr
      
      s       = quickcv.quickcv(side1, side2, np.array([zmid]), np.array([dz]), OmegaM = OmegaM, OmegaL = OmegaL, OmegaBaryon = OmegaBaryon, \
                                sigma8 = sigma8, ns = ns, h = h, acc = acc, verbose = verbose)
      
      ## These are the linear bias values
      bias70  = b070  * (zmid+1)**b170  + b270
      bias75  = b075  * (zmid+1)**b175  + b275
      bias80  = b080  * (zmid+1)**b180  + b280
      bias85  = b085  * (zmid+1)**b185  + b285
      bias90  = b090  * (zmid+1)**b190  + b290
      bias95  = b095  * (zmid+1)**b195  + b295
      bias100 = b0100 * (zmid+1)**b1100 + b2100
      bias105 = b0105 * (zmid+1)**b1105 + b2105
      bias110 = b0110 * (zmid+1)**b1110 + b2110

      ## All gg variables are galaxy-galaxy cosmic variance
      sgg70  = bias70  * s
      sgg75  = bias75  * s
      sgg80  = bias80  * s
      sgg85  = bias85  * s
      sgg90  = bias90  * s
      sgg95  = bias95  * s
      sgg100 = bias100 * s
      sgg105 = bias105 * s
      sgg110 = bias110 * s
      print(zmid, dz, s, sgg70, sgg75, sgg80, sgg85, sgg90, sgg95, sgg100, sgg105, sgg110)
      df_values.loc[i] = [zmid, dz, s, sgg70, sgg75, sgg80, sgg85, sgg90, sgg95, sgg100, sgg105, sgg110]
      #prints the mean redshift, redshift bins size, sigma_dm and the sigma_gg for nine stellar mass bins

   # Save the dataframes to csv files so that we don't have to run the code again
   if name:
      if not os.path.isdir('dfs'):
         os.mkdir('dfs')
      print('Saving dataframes of values to dfs/' + name )
      print(df_values)
      df_values.to_csv('dfs/' + name + '.csv', index = False)
      df_meta.to_csv('dfs/' + name + '_meta' + '.csv', index = False)
   return df_values

## Sample use
# if __name__ == '__main__':
#    get_cv(3.4/60, 10.2/60, np.array([3,4,5,6.5]), name = 'test')


import numpy as np

def pofkint(k0, k1, k2, OmegaM = 0.308, OmegaBaryon = 0.022/(0.678)**2, sigma8 = 0.82, ns = 0.96, h = 0.678):

    """ For 3d k, calculate a single k and returns the power spectrum

    OmegaM is the matter density parameter
    OmegaBaryon is the baryon density parameter, here it is set to the Planck value, 0.022/h^2
    sigma8 is the normalization of the power spectrum
    ns is the spectral index
    h is the hubble parameter H0/100

    returns the power at k = sqrt(k0**2 + k1**2 + k2**2)
    """

    k = np.linalg.norm(np.array([k0, k1, k2]))

    gamma = OmegaM*h*np.exp(-OmegaBaryon*(1.+np.sqrt(2*h)/OmegaM))

    keff=(0.172+0.011*(np.log(gamma/0.34))**2)*1.
    q=keff/(h*gamma)
    tk=np.log(1+2.34*q)/(2.34*q)*(1+3.89*q+(16.1*q)**2+(5.46*q)**3+(6.71*q)**4)**(-0.25)
    sigma8squn=keff**(3.0+ns)*tk**2

    q=k/(h*gamma)   # k/h Gamma
    tk=np.log(1+2.34*q)/(2.34*q)*(1+3.89*q+(16.1*q)**2+(5.46*q)**3+(6.71*q)**4)**(-0.25)
    delsq=k**ns*tk**2     
    delsq=delsq*(sigma8)**2/sigma8squn #normalize to Borgani et al. 

    pk=2*np.pi**2*delsq

    return pk


#!/usr/bin/env python

"""Tests for `cosmic_variance` package."""


import unittest

from cosmic_variance import cosmic_variance


class TestCosmic_variance(unittest.TestCase):
    """Tests for `cosmic_variance` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""



def Dlin(OmegaM, z): 
  '''Approximation for linear growth factor, assumes a flat cosmology, from Sean Carroll's paper
   OmegaM is the matter density parameter
   z is the redshift'''
  Omega = OmegaM*(1+z)**3/(1-OmegaM + (1+z)**3*OmegaM)
  gz = 2.5*Omega/(1./70. + 209.0*Omega/140. + Omega**(4./7.))
  g0 = 2.5*OmegaM/(1./70. + 209.0*OmegaM/140. + OmegaM**(4./7.))
  D = gz/(g0*(1+z))

  return D


"""Top-level package for Cosmic Variance."""

__author__ = """Christian Kragh Jespersen"""
__email__ = 'ckragh@princeton.edu'
__version__ = '0.1.2'

from .quickcv import quickcv
from .cosmic_variance import get_cv


#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['numpy>=1.17.0', 'scipy>=1.2.0', 'pandas>=1.0.0']

test_requirements = [ ]

setup(
    author="Christian Kragh Jespersen",
    author_email='ckragh@princeton.edu',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Package to calculate cosmic variance in rectangular pencil-beam surveys",
    install_requires=requirements,
    license="MIT license",
    long_description = readme,
    long_description_content_type='text/x-rst',
    include_package_data = True,
    keywords = ['Cosmology', 'Galaxies', 'Statistics', 'Astrostatistics', 'Cosmic Variance'],
    name = 'cosmic_variance',
    packages=find_packages(include=['cosmic_variance', 'cosmic_variance.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/astrockragh/cosmic_variance',
    version='0.1.2',
    zip_safe=False,
)
